# ///////////////////////////////////////////////////////////////
#
# BY: WANDERSON M.PIMENTA
# PROJECT MADE WITH: Qt Designer and PySide6
# V: 1.0.0
#
# This project can be used freely for all uses, as long as they maintain the
# respective credits only in the Python scripts, any information in the visual
# interface (GUI) can be modified without any implication.
#
# There are limitations on Qt licenses if you want to use your products
# commercially, I recommend reading them on the official website:
# https://doc.qt.io/qtforpython/licenses.html
#
# ///////////////////////////////////////////////////////////////
from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

# GUI FILE
from . ui_main import Ui_MainWindow

# APP SETTINGS
from . app_settings import Settings

# IMPORT FUNCTIONS
from . ui_functions import *

# APP FUNCTIONS
from . app_functions import *



import appdirs
import pathlib
import os
import yaml
from packaging import version as semantic_version
from . import __version__ as version

LOG_LEVEL = os.getenv("DJSCIOPS_LOG_LEVEL", "info")

DEFAULT_THRESHOLD = 64 * 1024**2  # 64 MiB
DEFAULT_CONCURRENCY = 100
DEFAULT_CHUNKSIZE = 64 * 1024**2  # 64 MiB
DEFAULT_THREAD_USE = True


def get_config(stdin_enabled=True):
    from .log import log

    config_directory = appdirs.user_data_dir(appauthor="datajoint", appname="djsciops")
    try:
        # loading existing config
        config = load_config(pathlib.Path(config_directory, "config.yaml").read_text())
        log.info(
            "Existing configuration detected. Loading from "
            f"{pathlib.Path(config_directory, 'config.yaml')}..."
        )
        # update djsciops config for boto3 transfer config addition
        if semantic_version.parse(config["version"]) < semantic_version.parse("1.2.0"):
            config["boto3"] = dict(
                multipart_threshold=DEFAULT_THRESHOLD,
                max_concurrency=DEFAULT_CONCURRENCY,
                multipart_chunksize=DEFAULT_CHUNKSIZE,
                use_threads=DEFAULT_THREAD_USE,
            )
            config["version"] = version
            save_config(yaml.dump(config), config_directory)
        # update djsciops config for issuer config addition
        if semantic_version.parse(config["version"]) < semantic_version.parse("1.5.0"):
            config["djauth"]["issuer"] = "https://accounts.datajoint.com/realms/master"
            config["version"] = version
            save_config(yaml.dump(config), config_directory)
    except FileNotFoundError as e:
        if not stdin_enabled:
            config = dict(
                version=version,
                aws=dict(account_id=""),
                s3=dict(role="", bucket=""),
                djauth=dict(client_id=""),
                boto3=dict(
                    multipart_threshold=DEFAULT_THRESHOLD,
                    max_concurrency=DEFAULT_CONCURRENCY,
                    multipart_chunksize=DEFAULT_CHUNKSIZE,
                    use_threads=DEFAULT_THREAD_USE,
                ),
            )
            log.warning("No config provided. Generating empty config.")
            return config
        log.info(
            "Welcome! We've detected that this is your first time using DataJoint "
            "SciOps CLI tools. We'll need to ask a few questions to initialize properly."
        )
        config = dict(
            version=version,
            aws=dict(
                account_id=os.getenv("DJSCIOPS_AWS_ACCOUNT_ID")
                or input("\n   -> AWS Account ID? ")
            ),
            s3=dict(
                role=os.getenv("DJSCIOPS_S3_ROLE") or input("\n   -> S3 Role? "),
                bucket=(
                    os.getenv("DJSCIOPS_S3_BUCKET") or input("\n   -> S3 Bucket? ")
                ),
            ),
            djauth=dict(
                issuer="https://accounts.datajoint.com/realms/master",
                client_id=(
                    os.getenv("DJ_AUTH_CLIENT_ID")
                    or input("\n   -> DataJoint Account Client ID? ")
                ),
                **(
                    dict(client_secret=secret)
                    if (secret := os.getenv("DJ_AUTH_CLIENT_SECRET"))
                    else {}
                ),
            ),
            boto3=dict(
                multipart_threshold=DEFAULT_THRESHOLD,
                max_concurrency=DEFAULT_CONCURRENCY,
                multipart_chunksize=DEFAULT_CHUNKSIZE,
                use_threads=DEFAULT_THREAD_USE,
            ),
        )
        # write config
        save_config(yaml.dump(config), config_directory)

        log.info(
            "Thank you! We've saved your responses to "
            f"{pathlib.Path(config_directory, 'config.yaml')} so you won't need to "
            "specify this again."
        )
        # return config
    return config


def load_config(config_str):
    from .log import log

    config = yaml.safe_load(config_str)
    if not config["aws"]["account_id"]:
        log.warning("Missing `aws.account_id` configuration")
    if not config["s3"]["role"]:
        log.warning("Missing `s3.role` configuration")
    if not config["s3"]["bucket"]:
        log.warning("Missing `s3.bucket` configuration")
    if not config["djauth"]["client_id"]:
        log.warning("Missing `djauth.client_id` configuration")
    return config


def save_config(config, config_directory):
    os.makedirs(config_directory, exist_ok=True)
    with open(pathlib.Path(config_directory, "config.yaml"), "w") as f:
        f.write(config)



import pathlib
import os
import boto3.s3.transfer
import tqdm
import re
from typing import Union
from djsciops import utils as djsciops_utils
from djsciops import settings as djsciops_settings
import collections.abc
import logging
from djsciops.log import log, TqdmToLogger
from datetime import datetime, timedelta
import time

_PERMIT_REGEX_DEFAULT = r".*"
_IGNORE_REGEX_DEFAULT = r"no-match^"
_DEFAULT_UPLOAD_MAPPING_RULES = ...
_DEFAULT_UPLOAD_MAPPING_INTERVAL = 60 * 60 * 6
_DEFAULT_UPLOAD_MAPPING_DURATION = None


def upload_continuously(
    session,
    s3_bucket: str,
    upload_mapping_config,
    boto3_config: dict = None,
    rules=_DEFAULT_UPLOAD_MAPPING_RULES,
    interval=_DEFAULT_UPLOAD_MAPPING_INTERVAL,
    duration=_DEFAULT_UPLOAD_MAPPING_DURATION,
    one_off=False,
):
    def _upload_single():
        source, destination = _generate_upload_mapping(
            upload_mapping_config=upload_mapping_config, rules=rules
        )
        upload_files(
            session=session,
            s3_bucket=s3_bucket,
            source=source,
            destination=destination,
            boto3_config=boto3_config,
        )

    if one_off:
        _upload_single()
    else:
        start_time = datetime.now()
        while (
            (datetime.now() <= start_time + timedelta(seconds=duration))
            if duration
            else True
        ):
            time.sleep(interval - time.time() % interval)
            _upload_single()


def _generate_upload_mapping(upload_mapping_config: dict, rules: list) -> tuple:
    rules = (
        upload_mapping_config["rules"]
        if rules is ...
        else [
            r
            for user_rule in rules
            for r in upload_mapping_config["rules"]
            if user_rule == r["name"]
        ]
    )
    upload_mapping = {
        pathlib.Path(path): pathlib.PurePosixPath(rule["destination_directory"])
        / pattern["destination_format"].format(**match.groupdict())
        for rule in list(reversed(rules))
        for pattern in list(reversed((rule["patterns"])))
        for path in pathlib.Path(rule["source_directory"]).glob("**/*")
        if (match := re.match(pattern["match_regex"], str(path)))
    }

    return list(upload_mapping), list(upload_mapping.values())


def _resolve_upload_mapping(
    session,
    s3_bucket: str,
    source: Union[str, list],
    destination: Union[str, list],
    *,
    permit_regex=_PERMIT_REGEX_DEFAULT,
    ignore_regex=_IGNORE_REGEX_DEFAULT,
) -> tuple:
    if (
        isinstance(source, list)
        and isinstance(destination, list)
        and len(source) == len(destination)
    ):
        log.debug("Upload mapping provided.")
        # expect file_paths to be list of Path, object_paths to be list of PurePosixPath
        file_paths, object_paths = (source, destination)
        existing_objects = {
            pathlib.PurePosixPath(op.key)
            for op in session.s3.Bucket(s3_bucket).objects.filter(
                Prefix=os.path.commonprefix(object_paths)
            )
        }
        results = tuple(
            zip(
                *(
                    (fp, op)
                    for fp, op in zip(file_paths, object_paths)
                    if (
                        op not in existing_objects
                        and re.match(permit_regex, fp.as_posix())
                        and not re.match(ignore_regex, fp.as_posix())
                    )
                )
            )
        )
        if not results:
            file_paths, object_paths = (None, None)
        else:
            file_paths, object_paths = results
    else:
        # modes: file -> directory, directory -> directory
        assert destination[-1] == "/", "Must point to a directory in object store."
        source = pathlib.Path(source).resolve()
        # recursively list files that are not hidden or directories
        file_paths = {
            relative_fp
            for fp in source.rglob("*")
            if (
                fp.is_file()
                and not fp.name.startswith(".")
                and re.match(
                    permit_regex,
                    (
                        relative_fp := pathlib.PurePosixPath(fp.relative_to(source))
                    ).as_posix(),
                )
                and not re.match(ignore_regex, relative_fp.as_posix())
            )
        }
        file_paths = file_paths or set(
            [pathlib.PurePosixPath(source.name)]
            if re.match(permit_regex, source.name)
            and not re.match(ignore_regex, source.name)
            else []
        )  # if specified a single file
        # recursively list objects
        existing_objects = {
            pathlib.PurePosixPath(op.key.replace(destination, ""))
            for op in session.s3.Bucket(s3_bucket).objects.filter(Prefix=destination)
        }
        # exclude objects that exist and verify that new objects are present
        file_paths = file_paths - existing_objects
        log.debug(f"file_paths: {file_paths}")
        if not file_paths:
            return [], []
        object_paths = [pathlib.PurePosixPath(destination, fp) for fp in file_paths]
        file_paths = [
            pathlib.Path(source if source.is_dir() else source.parent, fp)
            for fp in file_paths
        ]

    return file_paths, object_paths


def _resolve_download_mapping(
    session,
    s3_bucket: str,
    source: Union[str, list],
    destination: Union[str, list],
    *,
    permit_regex=_PERMIT_REGEX_DEFAULT,
    ignore_regex=_IGNORE_REGEX_DEFAULT,
) -> tuple:
    if (
        isinstance(source, list)
        and isinstance(destination, list)
        and len(source) == len(destination)
    ):
        log.debug("Download mapping provided.")
        # expect file_paths to be list of Path, object_paths to be list of PurePosixPath
        object_paths, file_paths = (source, destination)
        existing_files = {
            fp
            for fp in pathlib.Path(os.path.commonprefix(file_paths)).glob("**/*")
            if fp.is_file()
        }
        file_paths, object_paths = tuple(
            zip(
                *(
                    (fp, op)
                    for fp, op in zip(file_paths, object_paths)
                    if (
                        fp not in existing_files
                        and re.match(permit_regex, op.as_posix())
                        and not re.match(ignore_regex, op.as_posix())
                    )
                )
            )
        )

    else:
        # modes: file -> directory, directory -> directory
        assert destination[-1] == (
            "/" if os.name == "posix" else "\\"
        ), "Must point to a local directory"
        destination = pathlib.Path(destination).resolve()
        log.debug(f"destination: {destination}")
        # recursively list objects
        object_root = s.parent if (s := pathlib.PurePosixPath(source)).suffix else s
        object_paths = {
            relative_op
            for op in session.s3.Bucket(s3_bucket).objects.filter(Prefix=source)
            if (
                re.match(
                    permit_regex,
                    (
                        relative_op := pathlib.PurePosixPath(op.key).relative_to(
                            object_root
                        )
                    ).as_posix(),
                )
                and not re.match(ignore_regex, relative_op.as_posix())
            )
        }
        log.debug(f"object_paths: {object_paths}")
        # recursively list files that are not hidden or directories
        existing_objects = {
            fp.relative_to(destination)
            for fp in destination.rglob("*")
            if not fp.is_dir() and not str(fp.name).startswith(".")
        }
        log.debug(f"existing_objects: {existing_objects}")
        existing_objects = (
            {pathlib.PurePosixPath(_) for _ in existing_objects}
            if existing_objects
            else {pathlib.PurePosixPath(destination.name)}
        )  # if specified a single file
        # exclude objects that exist and verify that new objects are present
        object_paths = object_paths - existing_objects
        if not object_paths:
            return [], []
        file_paths = [pathlib.Path(destination, op) for op in object_paths]
        log.debug(f"file_paths: {file_paths}")
        object_paths = [pathlib.PurePosixPath(object_root, op) for op in object_paths]

    return object_paths, file_paths


def list_files(
    session,
    s3_bucket: str,
    s3_prefix: str,
    *,
    permit_regex=_PERMIT_REGEX_DEFAULT,
    ignore_regex=_IGNORE_REGEX_DEFAULT,
    include_contents_hash=False,
    as_tree=True,
):
    objects = [
        dict(
            key=op.key,
            _size=op.size,
            _contents_hash=op.Object().metadata["contents_hash"]
            if include_contents_hash
            else None,
        )
        for op in session.s3.Bucket(s3_bucket).objects.filter(Prefix=s3_prefix)
        if (
            re.match(permit_regex, op.key.replace(s3_prefix, ""))
            and not re.match(ignore_regex, op.key.replace(s3_prefix, ""))
        )
    ]
    objects = sorted(objects, key=lambda o: o["key"])

    if not as_tree:
        return objects

    tree = {}
    for o in objects:
        node = tree  # local node
        levels = o["key"].split("/")
        for level in levels:
            node = node.setdefault(level, dict())
            node["_size"] = (
                o["_size"] + node["_size"] if "_size" in node else o["_size"]
            )

    def update(d):
        for k, v in d.items():
            if isinstance(v, collections.abc.Mapping) and len(v) > 1:
                if k not in ("_size"):
                    _ = d[k].pop("_size")
                    d[k] = update(d.get(k, {}))
            else:
                d[k] = None
        return d

    return update(tree)


def upload_files(
    session,
    s3_bucket: str,
    source: Union[str, list],
    destination: Union[str, list],
    boto3_config: dict = None,
    *,
    permit_regex=_PERMIT_REGEX_DEFAULT,
    ignore_regex=_IGNORE_REGEX_DEFAULT,
):
    """
    Args:
        session: An instance of authentication.Session to access S3 resources
        s3_bucket: Name of the S3 bucket
        source: Source file or directory on client.
        destination: Target directory in object store.
        permit_regex: Regular expression pattern to permit file(s) for uploading (default - all files permitted)
        ignore_regex: Regular expression pattern to ignore file(s) for uploading (default - no files ignored)
    """
    if not boto3_config:
        boto3_config = djsciops_settings.get_config()["boto3"]

    file_paths, object_paths = _resolve_upload_mapping(
        session=session,
        s3_bucket=s3_bucket,
        source=source,
        destination=destination,
        permit_regex=permit_regex,
        ignore_regex=ignore_regex,
    )
    if not file_paths or not object_paths:
        log.warning("All upload-permitted files already exist in object store.")
        return
    log.debug("Starting upload")
    with tqdm.tqdm(
        total=sum(os.stat(fp).st_size for fp in file_paths),
        unit="B",
        unit_scale=True,
        desc="",
        mininterval=5,
        file=TqdmToLogger(log, level=logging.INFO),
    ) as pbar:
        for fp, op in zip(file_paths, object_paths):
            log.info(f"{fp}->{op}")
            ## hash metadata
            contents_hash = djsciops_utils.uuid_from_file(fp)
            log.debug(f"contents_hash.hex: {contents_hash.hex}")
            ## upload
            session.s3.Bucket(s3_bucket).upload_file(
                Filename=str(fp),
                Key=str(op),
                Config=boto3.s3.transfer.TransferConfig(**boto3_config),
                Callback=lambda bytes_transferred: pbar.update(bytes_transferred),
                ExtraArgs={"Metadata": {"contents_hash": contents_hash.hex}},
            )


def download_files(
    session,
    s3_bucket: str,
    source: Union[str, list],
    destination: Union[str, list],
    boto3_config: dict = None,
    *,
    permit_regex=_PERMIT_REGEX_DEFAULT,
    ignore_regex=_IGNORE_REGEX_DEFAULT,
):
    """

    Args:
        session: An instance of authentication.Session to access S3 resources
        s3_bucket: Name of the S3 bucket
        source: Source file or directory in object store.
        destination: Target directory on client.
        permit_regex: Regular expression pattern to permit file(s) for downloading (default - all files permitted)
        ignore_regex: Regular expression pattern to ignore file(s) for downloading (default - no files ignored)
    """

    if not boto3_config:
        boto3_config = djsciops_settings.get_config()["boto3"]

    object_paths, file_paths = _resolve_download_mapping(
        session=session,
        s3_bucket=s3_bucket,
        source=source,
        destination=destination,
        permit_regex=permit_regex,
        ignore_regex=ignore_regex,
    )
    if not file_paths or not object_paths:
        log.warning("All download-permitted files already exist locally.")
        return
    log.debug("Starting download")
    with tqdm.tqdm(
        total=sum(
            session.s3.Object(s3_bucket, str(op)).content_length for op in object_paths
        ),
        unit="B",
        unit_scale=True,
        desc="",
        mininterval=5,
        file=TqdmToLogger(log, level=logging.INFO),
    ) as pbar:
        for op, fp in zip(object_paths, file_paths):
            log.info(f"{op}->{fp}")
            # check if dir exists
            if not os.path.exists(os.path.dirname(fp)):
                os.makedirs(os.path.dirname(fp))
            ## download
            session.s3.Bucket(s3_bucket).download_file(
                Key=str(op),
                Filename=str(fp),
                Config=boto3.s3.transfer.TransferConfig(**boto3_config),
                Callback=lambda bytes_transferred: pbar.update(bytes_transferred),
            )



import base64
from datetime import datetime, timezone
import json
import logging
import flask
import webbrowser
import urllib
import http.client
import botocore
import botocore.config
from .log import log
from time import time

import boto3
from botocore.credentials import RefreshableCredentials
from botocore.session import get_session
from djsciops import settings as djsciops_settings

LOOKUP_SERVICE_ALLOWED_ORIGIN = "https://ops.datajoint.io"
LOOKUP_SERVICE_DOMAIN = "ops.datajoint.io"
LOOKUP_SERVICE_ROUTE = "/social-login/api/user"
LOOKUP_SERVICE_AUTH = {
    "https://accounts.datajoint.io/auth/": {
        "PROVIDER": "accounts.datajoint.io",
        "ROUTE": "/auth",
    },
    "https://accounts.datajoint.com/realms/master": {
        "PROVIDER": "accounts.datajoint.com",
        "ROUTE": "/realms/master/protocol/openid-connect",
    },
}
issuer = djsciops_settings.get_config()["djauth"]["issuer"]


def _client_login(
    auth_client_id: str,
    auth_client_secret: str,
    auth_provider_domain: str = LOOKUP_SERVICE_AUTH[issuer]["PROVIDER"],
    auth_provider_token_route: str = f"{LOOKUP_SERVICE_AUTH[issuer]['ROUTE']}/token",
):
    connection = http.client.HTTPSConnection(auth_provider_domain)
    headers = {"Content-type": "application/x-www-form-urlencoded"}
    body = urllib.parse.urlencode(
        {
            "grant_type": "client_credentials",
            "client_id": auth_client_id,
            "client_secret": auth_client_secret,
        }
    )
    connection.request("POST", auth_provider_token_route, body, headers)
    jwt_payload = json.loads(connection.getresponse().read().decode())
    return jwt_payload["access_token"]


def _oidc_login(
    auth_client_id: str,
    auth_url: str = f"https://{LOOKUP_SERVICE_AUTH[issuer]['PROVIDER']}{LOOKUP_SERVICE_AUTH[issuer]['ROUTE']}/auth",
    lookup_service_allowed_origin: str = LOOKUP_SERVICE_ALLOWED_ORIGIN,
    lookup_service_domain: str = LOOKUP_SERVICE_DOMAIN,
    lookup_service_route: str = LOOKUP_SERVICE_ROUTE,
    lookup_service_auth_provider: str = LOOKUP_SERVICE_AUTH[issuer]["PROVIDER"],
    code_challenge: str = "ubNp9Y0Y_FOENQ_Pz3zppyv2yyt0XtJsaPqUgGW9heA",
    code_challenge_method: str = "S256",
    code_verifier: str = "kFn5ZwL6ggOwU1OzKx0E1oZibIMC1ZbMC1WEUXcCV5mFoi015I9nB9CrgUJRkc3oiQT8uBbrvRvVzahM8OS0xJ51XdYaTdAlFeHsb6OZuBPmLD400ozVPrwCE192rtqI",
    callback_port: int = 28282,
    delay_seconds: int = 60,
):
    """
    Primary OIDC login flow.
    """
    import multiprocessing

    # Prepare user
    log.warning(
        "User authentication required to use DataJoint SciOps CLI tools. We'll be "
        "launching a web browser to authenticate your DataJoint account."
    )
    # allocate variables for access and context
    code = None
    cancelled = True
    # Prepare HTTP server to communicate with browser
    logging.getLogger("werkzeug").setLevel(logging.ERROR)
    app = flask.Flask("browser-interface")

    def shutdown_server():
        """
        Shuts down Flask HTTP server.
        """
        func = flask.request.environ.get("werkzeug.server.shutdown")
        if func is not None:
            # Ensure running with the Werkzeug Server
            func()

    @app.route("/login-cancelled")
    def login_cancelled():
        """
        Accepts requests which will cancel the user login.
        """
        shutdown_server()
        return """
        <!doctype html>
        <html>
          <head>
            <script>
              window.onload = function load() {
              window.open('', '_self', '');
              window.close();
              };
            </script>
          </head>
          <body>
          </body>
        </html>
        """

    @app.route("/login-completed")
    def login_completed():
        """
        Redirect after user has successfully logged in.
        """
        nonlocal code
        nonlocal cancelled
        cancelled = False
        code = flask.request.args.get("code")
        shutdown_server()
        return """
        <!doctype html>
        <html>
          <head>
            <script>
              window.onload = function load() {
              window.open('', '_self', '');
              window.close();
              };
            </script>
          </head>
          <body>DataJoint login completed! Feel free to close this tab if it did not close automatically.</body>
        </html>
        """

    # build url
    query_params = dict(
        scope="openid",
        response_type="code",
        client_id=auth_client_id,
        code_challenge=code_challenge,
        code_challenge_method=code_challenge_method,
        redirect_uri=f"http://localhost:{callback_port}/login-completed",
    )
    link = f"{auth_url}?{urllib.parse.urlencode(query_params)}"
    # attempt to launch browser or provide instructions
    browser_available = True
    try:
        webbrowser.get()
    except webbrowser.Error:
        browser_available = False
    if browser_available:
        log.info("Browser available. Launching...")
        webbrowser.open(link, new=2)
    else:
        log.warning(
            "Browser unavailable. On a browser client, please navigate to the "
            f"following link to login: {link}"
        )
    # start response server
    cancel_process = multiprocessing.Process(
        target=_delayed_request,
        kwargs=dict(
            url=f"http://localhost:{callback_port}/login-cancelled",
            delay=delay_seconds,
        ),
    )
    # cancel_process.start()
    app.run(host="0.0.0.0", port=callback_port, debug=False)
    # cancel_process.terminate()
    # received a response
    if cancelled:
        raise Exception(
            "User login cancelled. User must be logged in to use DataJoint SciOps CLI tools."
        )
    else:
        # generate user info
        connection = http.client.HTTPSConnection(lookup_service_domain)
        headers = {
            "Content-type": "application/json",
            "Origin": lookup_service_allowed_origin,
        }
        body = json.dumps(
            {
                "auth_provider": lookup_service_auth_provider,
                "redirect_uri": f"http://localhost:{callback_port}/login-completed",
                "code_verifier": code_verifier,
                "client_id": auth_client_id,
                "code": code,
            }
        )
        connection.request("POST", lookup_service_route, body, headers)
        userdata = json.loads(connection.getresponse().read().decode())
        log.info("User successfully authenticated.")
        return userdata["access_token"], userdata["username"], userdata["refresh_token"]


def _delayed_request(*, url: str, delay: str = 0):
    time.sleep(delay)
    return urllib.request.urlopen(url)


def _decode_bearer_token(bearer_token):
    log.debug(f"bearer_token: {bearer_token}")
    jwt_data = json.loads(
        base64.b64decode((bearer_token.split(".")[1] + "==").encode()).decode()
    )
    log.debug(f"jwt_data: {jwt_data}")
    return jwt_data


class Session:
    def __init__(
        self,
        aws_account_id: str,
        s3_role: str,
        auth_client_id: str,
        auth_client_secret: str = None,
        bearer_token: str = None,
    ):
        self.aws_account_id = aws_account_id
        self.s3_role = s3_role
        self.auth_client_id = auth_client_id
        self.auth_client_secret = auth_client_secret
        self.sts_arn = f"arn:aws:iam::{aws_account_id}:role/{s3_role}"
        self.user = "client_credentials"
        self.refresh_token = None
        self.jwt = None
        # OAuth2.0 authorization
        if auth_client_secret:
            self.bearer_token = _client_login(
                auth_client_id=self.auth_client_id,
                auth_client_secret=self.auth_client_secret,
            )
            self.jwt = _decode_bearer_token(self.bearer_token)
        elif not bearer_token:
            self.bearer_token, self.user, self.refresh_token = _oidc_login(
                auth_client_id=auth_client_id,
            )
            self.jwt = _decode_bearer_token(self.bearer_token)
        else:
            self.jwt = _decode_bearer_token(self.bearer_token)
            time_to_live = (self.jwt["exp"] - datetime.utcnow().timestamp()) / 60 / 60
            log.info(
                f"Reusing provided bearer token with a life of {time_to_live} [HR]"
            )
            self.bearer_token, self.user = (bearer_token, self.jwt["sub"])

        self.sts_token = RefreshableBotoSession(session=self).refreshable_session()
        self.s3 = self.sts_token.resource(
            "s3", config=botocore.config.Config(s3={"use_accelerate_endpoint": True})
        )

    def refresh_bearer_token(
        self,
        lookup_service_allowed_origin: str = LOOKUP_SERVICE_ALLOWED_ORIGIN,
        lookup_service_domain: str = LOOKUP_SERVICE_DOMAIN,
        lookup_service_route: str = LOOKUP_SERVICE_ROUTE,
        lookup_service_auth_provider: str = LOOKUP_SERVICE_AUTH[issuer]["PROVIDER"],
    ):
        if self.auth_client_secret:
            self.bearer_token = _client_login(
                auth_client_id=self.auth_client_id,
                auth_client_secret=self.auth_client_secret,
            )
            self.jwt = _decode_bearer_token(self.bearer_token)
        else:
            # generate user info
            connection = http.client.HTTPSConnection(lookup_service_domain)
            headers = {
                "Content-type": "application/json",
                "Origin": lookup_service_allowed_origin,
            }
            body = json.dumps(
                {
                    "auth_provider": lookup_service_auth_provider,
                    "refresh_token": self.refresh_token,
                    "client_id": self.auth_client_id,
                }
            )
            log.debug(f"Original refresh_token: {self.refresh_token}")
            connection.request("PATCH", lookup_service_route, body, headers)
            response = connection.getresponse().read().decode()
            log.debug(f"response: {response}")
            userdata = json.loads(response)
            log.debug("User successfully reauthenticated.")
            self.bearer_token = userdata["access_token"]
            self.user = userdata["username"]
            self.refresh_token = userdata["refresh_token"]
            log.debug(f"refresh_token: {self.refresh_token}")
            self.jwt = _decode_bearer_token(self.bearer_token)


class RefreshableBotoSession:
    """
    Boto Helper class which lets us create refreshable session, so that we can cache the client or resource.

    Usage
    -----
    session = RefreshableBotoSession().refreshable_session()

    client = session.client("s3") # we now can cache this client object without worrying about expiring credentials
    """

    def __init__(self, session, session_ttl: int = 12 * 60 * 60):
        """
        Initialize `RefreshableBotoSession`

        Parameters
        ----------
        session : Session
            The session object to refresh

        session_ttl : int (optional)
            An integer number to set the TTL for each session. Beyond this session, it will renew the token.
        """

        self.session = session
        self.session_ttl = session_ttl

    def __get_session_credentials(self):
        """
        Get session credentials
        """
        sts_client = boto3.client(service_name="sts")
        try:
            sts_response = sts_client.assume_role_with_web_identity(
                RoleArn=self.session.sts_arn,
                RoleSessionName=self.session.user,
                WebIdentityToken=self.session.bearer_token,
                DurationSeconds=self.session_ttl,
            ).get("Credentials")
        except botocore.exceptions.ClientError as error:
            log.debug(f"Error code: {error.response['Error']['Code']}")
            if error.response["Error"]["Code"] == "ExpiredTokenException":
                log.debug("Bearer token has expired... Reauthenticating now")
                self.session.refresh_bearer_token()
                sts_response = sts_client.assume_role_with_web_identity(
                    RoleArn=self.session.sts_arn,
                    RoleSessionName=self.session.user,
                    WebIdentityToken=self.session.bearer_token,
                    DurationSeconds=self.session_ttl,
                ).get("Credentials")
            else:
                raise error
        # Token expire time logging
        bearer_expire_time = datetime.fromtimestamp(self.session.jwt["exp"]).strftime(
            "%H:%M:%S"
        )
        log.debug(f"Bearer token expire time: {bearer_expire_time}")
        if "sts_token" in self.session.__dict__:
            sts_expire_time = (
                self.session.sts_token._session.get_credentials()
                .__dict__["_expiry_time"]
                .replace(tzinfo=timezone.utc)
                .astimezone(tz=None)
                .strftime("%H:%M:%S")
            )
            log.debug(f"STS token expire time: {sts_expire_time}")

        credentials = {
            "access_key": sts_response.get("AccessKeyId"),
            "secret_key": sts_response.get("SecretAccessKey"),
            "token": sts_response.get("SessionToken"),
            "expiry_time": sts_response.get("Expiration").isoformat(),
        }

        return credentials

    def refreshable_session(self) -> boto3.Session:
        """
        Get refreshable boto3 session.
        """
        # get refreshable credentials
        refreshable_credentials = RefreshableCredentials.create_from_metadata(
            metadata=self.__get_session_credentials(),
            refresh_using=self.__get_session_credentials,
            method="sts-assume-role-with-web-identity",
        )

        # attach refreshable credentials current session
        session = get_session()
        session._credentials = refreshable_credentials
        autorefresh_session = boto3.Session(botocore_session=session)

        return autorefresh_session



import argparse
import sys
import yaml
from djsciops import __version__ as version
from djsciops import authentication as djsciops_authentication
from djsciops import axon as djsciops_axon
from djsciops import settings as djsciops_settings
from djsciops.axon import speed_audit as djsciops_speed_audit


def djsciops(args: list = None):
    """
    Primary console interface for djsciops's shell utilities.

    :param args: List of arguments to be passed in, defaults to reading stdin
    :type args: list, optional
    """
    from djsciops.log import log

    parser = argparse.ArgumentParser(
        prog="djsciops", description="DataJoint SciOps console interface."
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"djsciops {version}"
    )
    command = parser.add_subparsers(dest="command")

    axon = command.add_parser("axon", description="Manage object store data.")
    axon_subcommand = axon.add_subparsers(dest="subcommand")

    axon_upload_agent = axon_subcommand.add_parser(
        "upload-agent",
        description="Upload objects in the background.",
    )
    axon_upload_agent_optional_named = axon_upload_agent.add_argument_group(
        "optional named arguments"
    )
    axon_upload_agent_optional_named.add_argument(
        "--one-off",
        action="store_true",
        dest="one_off",
        help="Upload mapping rules only once.",
    )
    axon_upload_agent_optional_named.add_argument(
        "rules",
        type=str,
        nargs="*",
        default=djsciops_axon._DEFAULT_UPLOAD_MAPPING_RULES,
        help="List of specific rules and check order to perform upload. Defaults to using all rules.",
    )
    axon_upload_agent_optional_named.add_argument(
        "-i",
        "--interval",
        type=int,
        default=djsciops_axon._DEFAULT_UPLOAD_MAPPING_INTERVAL,
        required=False,
        help="Time between every upload in seconds. Default to 21,600 (6 hours).",
    )
    axon_upload_agent_optional_named.add_argument(
        "-d",
        "--duration",
        type=int,
        default=djsciops_axon._DEFAULT_UPLOAD_MAPPING_DURATION,
        required=False,
        help="Stop uploading after a certain amount of time. Default to uploading indefinitely.",
    )

    axon_upload = axon_subcommand.add_parser(
        "upload", description="Copy objects by uploading to object store."
    )
    axon_upload_required_named = axon_upload.add_argument_group(
        "required named arguments"
    )
    axon_upload_required_named.add_argument(
        "source",
        type=str,
        help="Source file or directory on client.",
    )
    axon_upload_required_named.add_argument(
        "destination",
        type=str,
        help="Target directory in object store.",
    )
    axon_upload_optional_named = axon_upload.add_argument_group(
        "optional named arguments"
    )
    axon_upload_optional_named.add_argument(
        "-p",
        "--permit-regex",
        type=str,
        default=djsciops_axon._PERMIT_REGEX_DEFAULT,
        required=False,
        dest="permit_regex",
        help="Regular expression pattern to permit file(s) for uploading (default - all files permitted).",
    )
    axon_upload_optional_named.add_argument(
        "-i",
        "--ignore-regex",
        type=str,
        default=djsciops_axon._IGNORE_REGEX_DEFAULT,
        required=False,
        dest="ignore_regex",
        help="Regular expression pattern to ignore file(s) for uploading (default - no files ignored).",
    )

    axon_download = axon_subcommand.add_parser(
        "download", description="Download objects from object store."
    )
    axon_download_required_named = axon_download.add_argument_group(
        "required named arguments"
    )
    axon_download_required_named.add_argument(
        "source",
        type=str,
        help="Source file or directory in object store.",
    )
    axon_download_required_named.add_argument(
        "destination",
        type=str,
        help="Target directory on client.",
    )

    axon_download_optional_named = axon_download.add_argument_group(
        "optional named arguments"
    )
    axon_download_optional_named.add_argument(
        "-p",
        "--permit-regex",
        type=str,
        default=djsciops_axon._PERMIT_REGEX_DEFAULT,
        required=False,
        dest="permit_regex",
        help="Regular expression pattern to permit file(s) for downloading (default - all files permitted).",
    )
    axon_download_optional_named.add_argument(
        "-i",
        "--ignore-regex",
        type=str,
        default=djsciops_axon._IGNORE_REGEX_DEFAULT,
        required=False,
        dest="ignore_regex",
        help="Regular expression pattern to ignore file(s) for downloading (default - no files ignored).",
    )

    axon_speed_audit = axon_subcommand.add_parser(
        "speed-audit",
        description="Compare the speed of axon to the speed of the network.",
    )

    axon_speed_audit_optional_named = axon_speed_audit.add_argument_group(
        "optional named arguments"
    )
    axon_speed_audit_optional_named.add_argument(
        "-s",
        "--file-size",
        type=int,
        default=djsciops_speed_audit._DEFAULT_FILE_SIZE,
        required=False,
        dest="file_size",
        help=f"File size in MiB of generated file used during test. Defaults to {djsciops_speed_audit._DEFAULT_FILE_SIZE}.",
    )
    axon_speed_audit_optional_named.add_argument(
        "-d",
        "--down-limit",
        type=int,
        default=djsciops_speed_audit._DEFAULT_DOWN_LIMIT,
        required=False,
        dest="down_limit",
        help="Expected download bandwidth in Mbps. Defaults to skipping comparison.",
    )
    axon_speed_audit_optional_named.add_argument(
        "-u",
        "--up-limit",
        type=int,
        default=djsciops_speed_audit._DEFAULT_UP_LIMIT,
        required=False,
        dest="up_limit",
        help="Expected upload bandwidth in Mbps. Defaults to skipping comparison.",
    )

    axon_speed_audit_optional_named.add_argument(
        "-p",
        "--path",
        type=str,
        default=djsciops_speed_audit._DEFAULT_PATH,
        required=False,
        dest="path",
        help="Audit from a specific directory. Defaults to using local temporary directory.",
    )

    config = command.add_parser("config", description="View or modify djsciops config")

    config_optional_named = config.add_argument_group("optional named arguments")
    config_optional_named.add_argument(
        "key", type=str, nargs="?", default=None, help="Configuration key."
    )
    config_optional_named.add_argument(
        "value", type=str, nargs="?", default=None, help="New configuration value."
    )
    kwargs = vars(parser.parse_args(args if sys.argv[1:] else ["-h"]))
    djsciops_config = djsciops_settings.get_config()
    if kwargs["command"] == "axon":
        if kwargs["subcommand"] == "upload-agent":
            djsciops_axon.upload_continuously(
                one_off=kwargs["one_off"],
                session=djsciops_authentication.Session(
                    aws_account_id=djsciops_config["aws"]["account_id"],
                    s3_role=djsciops_config["s3"]["role"],
                    auth_client_id=djsciops_config["djauth"]["client_id"],
                    **(
                        {"auth_client_secret": c["client_secret"]}
                        if "client_secret" in (c := djsciops_config["djauth"])
                        else {}
                    ),
                ),
                upload_mapping_config=djsciops_config["axon"]["upload"],
                s3_bucket=djsciops_config["s3"]["bucket"],
                boto3_config=djsciops_config["boto3"],
                rules=kwargs["rules"],
                interval=kwargs["interval"],
                duration=kwargs["duration"],
            )
        elif kwargs["subcommand"] == "upload":
            djsciops_axon.upload_files(
                session=djsciops_authentication.Session(
                    aws_account_id=djsciops_config["aws"]["account_id"],
                    s3_role=djsciops_config["s3"]["role"],
                    auth_client_id=djsciops_config["djauth"]["client_id"],
                    **(
                        {"auth_client_secret": c["client_secret"]}
                        if "client_secret" in (c := djsciops_config["djauth"])
                        else {}
                    ),
                ),
                s3_bucket=djsciops_config["s3"]["bucket"],
                source=kwargs["source"],
                destination=kwargs["destination"],
                boto3_config=djsciops_config["boto3"],
                permit_regex=kwargs["permit_regex"],
                ignore_regex=kwargs["ignore_regex"],
            )
        elif kwargs["subcommand"] == "download":
            djsciops_axon.download_files(
                session=djsciops_authentication.Session(
                    aws_account_id=djsciops_config["aws"]["account_id"],
                    s3_role=djsciops_config["s3"]["role"],
                    auth_client_id=djsciops_config["djauth"]["client_id"],
                    **(
                        {"auth_client_secret": c["client_secret"]}
                        if "client_secret" in (c := djsciops_config["djauth"])
                        else {}
                    ),
                ),
                s3_bucket=djsciops_config["s3"]["bucket"],
                source=kwargs["source"],
                destination=kwargs["destination"],
                boto3_config=djsciops_config["boto3"],
                permit_regex=kwargs["permit_regex"],
                ignore_regex=kwargs["ignore_regex"],
            )
        elif kwargs["subcommand"] == "speed-audit":
            djsciops_speed_audit.speed_audit(
                **dict(
                    {k: v for k, v in kwargs.items() if "command" not in k},
                    aws_account_id=djsciops_config["aws"]["account_id"],
                    s3_bucket=djsciops_config["s3"]["bucket"],
                    region="us-east-2",
                    boto3_config=djsciops_config["boto3"],
                )
            )
    elif kwargs["command"] == "config":

        def _recursive_index(key, config_obj, mode="get"):
            if "." not in key:
                if mode == "get":
                    return config_obj[key]
                elif mode == "set":
                    return key, config_obj
            config_keys = key.split(".")
            return _recursive_index(
                ".".join(config_keys[1:]), config_obj[config_keys[0]], mode=mode
            )

        def _cast_input(s, key):
            if s.lower() == "true":
                return True
            elif s.lower() == "false":
                return False
            elif not s.isnumeric() or key == "aws.account_id":
                return s
            else:
                try:
                    return int(s)
                except ValueError:
                    return float(s)

        if kwargs["key"] and not kwargs["value"]:
            log.info(
                yaml.dump(_recursive_index(kwargs["key"], djsciops_config)),
                extra={"disable_format": True},
            )

        elif kwargs["key"] and kwargs["value"]:
            curr_key, curr_object = _recursive_index(
                kwargs["key"], djsciops_config, mode="set"
            )
            curr_object[curr_key] = _cast_input(kwargs["value"], kwargs["key"])
            config_directory = djsciops_settings.appdirs.user_data_dir(
                appauthor="datajoint", appname="djsciops"
            )
            djsciops_settings.save_config(yaml.dump(djsciops_config), config_directory)

        else:
            log.info(yaml.dump(djsciops_config), extra={"disable_format": True})
    raise SystemExit


if __name__ == "__main__":
    djsciops()



import pathlib
import tempfile
import time
from datetime import datetime
import speedtest
from djsciops import authentication
from djsciops import axon
from djsciops.log import log
from djsciops import settings as djsciops_settings
from minio import Minio
from minio.credentials import WebIdentityProvider

_DEFAULT_FILE_SIZE = 256
_DEFAULT_UP_LIMIT = None
_DEFAULT_DOWN_LIMIT = None
_DEFAULT_PATH = None


def speed_audit(
    aws_account_id: str,
    s3_bucket: str,
    region: str,
    boto3_config: dict,
    file_size: int = _DEFAULT_FILE_SIZE,
    up_limit: int = _DEFAULT_UP_LIMIT,
    down_limit: int = _DEFAULT_DOWN_LIMIT,
    path: str = _DEFAULT_PATH,
):
    def perform_network_test(description: str, downloads: list, uploads: list) -> tuple:
        log.info(f"Starting {description} network test...")
        return downloads + [st.download() / 10**6], uploads + [st.upload() / 10**6]

    def time_it(function):
        start_time = time.time()
        function()
        return time.time() - start_time

    def display(direction, unit, network, s3, axon, bandwidth=None):
        custom_round = lambda x: x if isinstance(x, str) else round(x, 2)
        constant = 100 / bandwidth if bandwidth else 1
        network_string = (
            f"Network: {custom_round(network*constant)} {unit}, " if network else ""
        )
        log.info(
            f"{direction} => {network_string}S3: {custom_round(s3*constant)} {unit}, Axon: {custom_round(axon*constant)} {unit}"
        )

    log.info("Initializing test...")
    session_data = {
        "aws_account_id": aws_account_id,
        "s3_role": "axon_speed_audit",
        "auth_client_id": "axon-speed-audit",
        "auth_client_secret": "axon-speed-audit",
    }
    s3_directory = pathlib.PurePosixPath("test/axon/speed-audit")
    name = "speed-audit-{size}{suffix}.tmp"
    remote_path = s3_directory / name.format(
        size=file_size, suffix=f"-{int(datetime.utcnow().timestamp())}"
    )
    local_path = pathlib.Path(path or f"{tempfile.gettempdir()}/axon") / name.format(
        size=file_size, suffix=""
    )
    log.debug(f"local_path: {local_path}")
    if not path:
        local_path.parent.mkdir(parents=True, exist_ok=True)
    byte_length = file_size * 1024**2
    if local_path.is_file():
        log.debug("Found test file. Skipping generating file.")
    else:
        log.debug(f"Generating {file_size} MiB file...")
        with open(local_path, "wb") as f:
            f.seek(byte_length - 1)
            f.write(("\0").encode())
    session = authentication.Session(**session_data)
    provider = WebIdentityProvider(
        jwt_provider_func=lambda: {"access_token": session.bearer_token},
        sts_endpoint="https://sts.amazonaws.com:443",
        role_arn=f"arn:aws:iam::{session_data['aws_account_id']}:role/{session_data['s3_role']}",
    )
    client = Minio(
        endpoint="s3-accelerate.amazonaws.com:443", region=region, credentials=provider
    )
    download, upload = ([], [])
    try:
        st = speedtest.Speedtest()
        st.get_best_server()
        enable_network_test = True
    except speedtest.ConfigRetrievalError:
        log.warning("Network speed test temporarily unavailable!")
        enable_network_test = False
    # Run tests
    if enable_network_test:
        download, upload = perform_network_test("first", download, upload)
    log.info("Running axon upload test...")
    axon_up_time = time_it(
        lambda: axon.upload_files(
            session, s3_bucket, [local_path], [remote_path], boto3_config
        )
    )
    log.info("Running S3 upload test...")
    s3_up_time = time_it(
        lambda: client.fput_object(s3_bucket, str(remote_path), str(local_path))
    )
    local_path.unlink()
    if enable_network_test:
        download, upload = perform_network_test("second", download, upload)
    log.info("Starting S3 download test...")
    s3_down_time = time_it(
        lambda: client.fget_object(s3_bucket, str(remote_path), str(local_path))
    )
    log.info("Starting axon download test...")
    axon_down_time = time_it(
        lambda: axon.download_files(
            session, s3_bucket, [remote_path], [local_path], boto3_config
        )
    )
    if enable_network_test:
        download, upload = perform_network_test("final", download, upload)
    # Summarize results
    net_up_speed = sum(upload) / len(upload) if enable_network_test else None
    net_down_speed = sum(download) / len(download) if enable_network_test else None
    s3_up_speed = byte_length * 8 / 10**6 / s3_up_time
    s3_down_speed = byte_length * 8 / 10**6 / s3_down_time
    axon_up_speed = byte_length * 8 / 10**6 / axon_up_time
    axon_down_speed = byte_length * 8 / 10**6 / axon_down_time
    if up_limit:
        display("Upload", "%", net_up_speed, s3_up_speed, axon_up_speed, up_limit)
    if down_limit:
        display(
            "Download", "%", net_down_speed, s3_down_speed, axon_down_speed, down_limit
        )
    display("Upload", "Mbps", net_up_speed, s3_up_speed, axon_up_speed)
    display("Download", "Mbps", net_down_speed, s3_down_speed, axon_down_speed)
